//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002-2004
//	Sleepycat Software.  All rights reserved.
//
// $Id: DbXmlURIResolver.cpp,v 1.7 2004/12/03 03:01:01 gmf Exp $
//

#include <dbxml/XmlValue.hpp>
#include <dbxml/XmlResults.hpp>
#include <dbxml/XmlResolver.hpp>
#include <dbxml/XmlTransaction.hpp>
#include <dbxml/XmlData.hpp>
#include "DbXmlURIResolver.hpp"
#include "DbXmlContext.hpp"
#include "DbXmlUri.hpp"
#include "../UTF8.hpp"
#include "../Value.hpp"
#include "../Results.hpp"
#include "../Manager.hpp"
#include "../XmlInputStreamWrapper.hpp"
#include "../nodeStore/NsUtil.hpp"

#include <pathan/Sequence.hpp>
#include <pathan/exceptions/XMLParseException.hpp>

#include <xercesc/framework/XMLBuffer.hpp>
#include <xercesc/util/RuntimeException.hpp>
#include <xercesc/util/XMLURL.hpp>

#if defined(XERCES_HAS_CPP_NAMESPACE)
XERCES_CPP_NAMESPACE_USE
#endif

using namespace DbXml;
using namespace std;

XmlInputStream *DbXmlSchemeResolver::resolveSchema(
	XmlTransaction *txn, XmlManager &mgr,
	const std::string &schemaLocation,
	const std::string &nameSpace) const
{
	DbXmlUri uri(schemaLocation, /*documentUri*/true);
	if(uri.isValid()) {
		Transaction *t = 0;
		if (txn)
			t = *txn;
		XmlDocument doc(uri.openDocument(mgr, t));
		// Can't use getContentAsInputStream(), as the XmlInputStream
		// returned does not hold a reference to the XmlDocument, and
		// therefore becomes invalid when the XmlDocument goes out of
		// scope - jpcs
		XmlData content(doc.getContent());
		char *copy = new char[content.get_size()];
		memcpy(copy, content.get_data(), content.get_size());
		return mgr.createMemBufInputStream(copy, content.get_size(),
						   schemaLocation.c_str(),
						   /*adoptBuffer*/true);
	}
	return 0;
}

XmlInputStream *DbXmlSchemeResolver::resolveEntity(
	XmlTransaction *txn, XmlManager &mgr,
	const std::string &systemId,
	const std::string &publicId) const
{
	DbXmlUri uri(systemId, /*documentUri*/true);
	if(uri.isValid()) {
		Transaction *t = 0;
		if (txn)
			t = *txn;
		XmlDocument doc(uri.openDocument(mgr, t));
		// Can't use getContentAsInputStream(), as the XmlInputStream
		// returned does not hold a reference to the XmlDocument, and
		// therefore becomes invalid when the XmlDocument goes out of
		// scope - jpcs
		XmlData content(doc.getContent());
		char *copy = new char[content.get_size()];
		memcpy(copy, content.get_data(), content.get_size());
		return mgr.createMemBufInputStream(copy, content.get_size(),
						   systemId.c_str(),
						   /*adoptBuffer*/true);
	}
	return 0;
}

ResolverStore::ResolverStore()
	: secure_(false)
{
 	registerResolver(dbxmlResolver_);
}

void ResolverStore::registerResolver(const XmlResolver &resolver)
{
	resolvers_.push_back(&resolver);
}

ResolverStore::const_iterator ResolverStore::begin() const
{
	return resolvers_.begin();
}

ResolverStore::const_iterator ResolverStore::end() const
{
	return resolvers_.end();
}

DbXmlURIResolver::DbXmlURIResolver(XmlManager &mgr, Transaction *txn)
	: mgr_(mgr),
	  txn_(txn)
{
}

static XMLCh *_createURI(const XMLCh *base, const XMLCh *uri)
{
	XMLCh *ret;
	int len = NsUtil::nsStringLen(uri);
	int baseLen = 0;
	if (base)
		baseLen = NsUtil::nsStringLen(base);
	ret = (XMLCh*) ::malloc((baseLen + len + 1) * sizeof(XMLCh));
	if (!ret)
		throw XmlException(XmlException::NO_MEMORY_ERROR,
				   "Cannot allocate memory for uri");
	if (base)
		::memcpy(ret, base, baseLen * sizeof(XMLCh));
	::memcpy(&(ret[baseLen]), uri, len * sizeof(XMLCh));
	ret[baseLen + len] = 0;
	// trust the result, since setBaseURI validates the base
	// as having the form, xxxx:*
	return ret;
}

static void _releaseURI(XMLCh *str)
{
	::free(str);
}

bool DbXmlURIResolver::resolveDocument(Sequence &result, const XMLCh* uri, DynamicContext* context)
{
	DbXmlContext *dbContext = CAST_TO_DBXMLCONTEXT(context);

	QueryContext &qc = dbContext->getQueryContext();

	// Concat base URI (if present) and uri
	// TBD: consider using XMLUri class to validate the result
	XMLCh *systemId = _createURI(context->getBaseURI(), uri);
	bool retVal = true;
	try {
		XmlValue value;
		if(resolveDocument(XMLChToUTF8(systemId).str(), value)) {
			result.addItem(Value::convertToItem(value, context));
		} else if(((Manager&)mgr_).getResolverStore().getSecure()) {
			// If we're secure, we don't let Pathan
			// run it's default lookup
			XMLBuffer errMsg;
			errMsg.set(UTF8ToXMLCh("Error retrieving resource: ").
				   str());
			errMsg.append(systemId);
			DSLthrow(XMLParseException,
				 UTF8ToXMLCh("DbXmlURIResolver::resolveDocument").str(),
				 errMsg.getRawBuffer());
		} else {
			retVal = false;
		}
	} catch (...) {
		_releaseURI(systemId);
		throw; // re-throw
	}
	return retVal;
}

bool DbXmlURIResolver::resolveCollection(Sequence &result, const XMLCh* uri, DynamicContext* context)
{
	DbXmlContext *dbContext = CAST_TO_DBXMLCONTEXT(context);

	QueryContext &qc = dbContext->getQueryContext();

	// Concat base URI (if present) and uri
	// TBD: consider using XMLUri class to validate the result
	XMLCh *systemId = _createURI(context->getBaseURI(), uri);
	bool retVal = true;
	try {
		XmlResults results = new ValueResults();
		if(resolveCollection(XMLChToUTF8(systemId).str(), results)) {
			results.reset();
			XmlValue value;
			while(results.next(value)) {
				result.addItem(Value::convertToItem(value,
								    context));
			}
		} else if(((Manager&)mgr_).getResolverStore().getSecure()) {
			// If we're secure, we don't let Pathan run
			// it's default lookup
			XMLBuffer errMsg;
			errMsg.set(UTF8ToXMLCh("Error retrieving resource: ").
				   str());
			errMsg.append(systemId);
			DSLthrow(XMLParseException,
				 UTF8ToXMLCh("DbXmlURIResolver::resolveCollection").str(), errMsg.getRawBuffer());
		} else {
			retVal = false;
		}
	} catch (...) {
		_releaseURI(systemId);
		throw; // re-throw
	}
	return retVal;
}

InputSource* DbXmlURIResolver::resolveEntity(XMLResourceIdentifier *ri)
{
	XmlInputStream *result = 0;
	switch(ri->getResourceIdentifierType()) {
	case XMLResourceIdentifier::SchemaGrammar:
	case XMLResourceIdentifier::SchemaImport:
	case XMLResourceIdentifier::SchemaInclude:
	case XMLResourceIdentifier::SchemaRedefine: {
		result = resolveSchema(XMLChToUTF8(ri->getSchemaLocation()).str(),
				       XMLChToUTF8(ri->getNameSpace()).str());
		break;
	}
	case XMLResourceIdentifier::ExternalEntity: {
		result = resolveEntity(XMLChToUTF8(ri->getSystemId()).str(),
				       XMLChToUTF8(ri->getPublicId()).str());
		break;
	}
	default: {
		// Stop compiler warnings
		break;
	}
	}

	if(result != 0) {
		return new XmlInputStreamWrapper(&result);
	}
	else if(((Manager&)mgr_).getResolverStore().getSecure()) {
		ThrowXML(RuntimeException, XMLExcepts::Gen_CouldNotOpenExtEntity);
	}
	else {
		return 0;
	}

}

bool DbXmlURIResolver::resolveDocument(const std::string &uri,
				       XmlValue &result) const
{
	scoped_ptr<XmlTransaction> txn;
	if(txn_) txn.reset(new XmlTransaction(txn_));

	ResolverStore::const_iterator end = ((Manager&)mgr_).getResolverStore().end();
	for(ResolverStore::const_iterator it = ((Manager&)mgr_).getResolverStore().begin(); it != end; ++it) {
		if((*it)->resolveDocument(txn.get(), mgr_, uri, result)) {
			return true;
		}
	}
	return false;
}

bool DbXmlURIResolver::resolveCollection(const std::string &uri,
					 XmlResults &result) const
{
	scoped_ptr<XmlTransaction> txn;
	if(txn_) txn.reset(new XmlTransaction(txn_));

	ResolverStore::const_iterator end = ((Manager&)mgr_).getResolverStore().end();
	for(ResolverStore::const_iterator it = ((Manager&)mgr_).getResolverStore().begin(); it != end; ++it) {
		if((*it)->resolveCollection(txn.get(), mgr_, uri, result)) {
			return true;
		}
	}
	return false;
}

XmlInputStream *DbXmlURIResolver::resolveSchema(
	const std::string &schemaLocation, const std::string &nameSpace) const
{
	scoped_ptr<XmlTransaction> txn;
	if(txn_) txn.reset(new XmlTransaction(txn_));

	XmlInputStream *result = 0;
	ResolverStore::const_iterator end = ((Manager&)mgr_).getResolverStore().end();
	for(ResolverStore::const_iterator it = ((Manager&)mgr_).getResolverStore().begin(); it != end; ++it) {
		result = (*it)->resolveSchema(txn.get(), mgr_, schemaLocation, nameSpace);
		if(result != 0) break;
	}
	return result;
}

XmlInputStream *DbXmlURIResolver::resolveEntity(
	const std::string &systemId, const std::string &publicId) const
{
	scoped_ptr<XmlTransaction> txn;
	if(txn_) txn.reset(new XmlTransaction(txn_));

	XmlInputStream *result = 0;
	ResolverStore::const_iterator end =
		((Manager&)mgr_).getResolverStore().end();
	for(ResolverStore::const_iterator it =
		    ((Manager&)mgr_).getResolverStore().begin(); it != end; ++it) {
		result = (*it)->resolveEntity(txn.get(), mgr_, systemId,
					      publicId);
		if(result != 0) break;
	}
	return result;
}
